package net.cattaka.slidepuzzle.solver;
import java.util.ArrayList;
import java.util.List;

public class SolveStage {
	private byte[] stage = new byte[36];
	private byte wx;
	private byte wy;
	private int maxLevel;
	private int maxScore;
	private List<List<List<Integer>>> connectedGroup;

	public SolveStage(byte[] tile, byte wx, byte wy) {
		this.stage = new byte[tile.length];
		this.wx = wx;
		this.wy = wy;
		createStage(tile);
	}

	private void createStage(byte[] tile) {
		int level = 0;
		while (true) {
			boolean goflag = false;
			maxLevel = level;
			level++;
			byte[] tStage = new byte[stage.length];
			System.arraycopy(stage, 0, tStage, 0, stage.length);

			for (int x = 0; x < wx; x++) {
				for (int y = 0; y < wy; y++) {
					int rn = getRouteNum(tile, tStage, level, x, y);
					if (rn == -1) {
						// none
					} else if (rn <= 2) {
						goflag = true;
						stage[x + wx * y] = (byte) level;
					}
				}
			}
			if (!goflag) {
				break;
			}
			while (true) {
				System.arraycopy(stage, 0, tStage, 0, stage.length);
				goflag = false;

				for (int x = 0; x < wx; x++) {
					for (int y = 0; y < wy; y++) {
						int rn = getRouteNum(tile, tStage, level, x, y);
						if (rn == -1) {
							// none
						} else if (rn < 2) {
							goflag = true;
							stage[x + wx * y] = (byte) level;
						}
					}
				}
				if (!goflag) {
					break;
				}
			}
			//System.out.println(this.toString());
		}
		updateEndEdge();
		connectedGroup = pickConnectedGroup();
		maxScore = 0;
		for (List<List<Integer>> groupIdsList : connectedGroup) {
			maxScore += groupIdsList.size();
		}
	}

	private void updateEndEdge() {
		List<Integer> edge = new ArrayList<Integer>();
		{
			boolean[] picked = new boolean[stage.length];
			int level = stage[stage.length - 1];
			pickConnectedGroup(edge, picked, level, wx - 1, wy - 1);
		}
		do {
			List<Integer> result = new ArrayList<Integer>();
			boolean[] picked = new boolean[stage.length];
			int level = stage[stage.length - 1];
			if (level == maxLevel) {
				break;
			}
			pickConnectedGroup(result, picked, level, wx - 1, wy - 1);
			for (Integer i : result) {
				stage[i] = (byte) (level + 1);
			}
		} while (true);
		for (Integer i : edge) {
			stage[i] = (byte) (maxLevel+1);
		}
	}

	private int getRouteNum(byte[] tile, byte[] tStage, int level, int x, int y) {
		if (isStageWall(tile, tStage, level, x, y)) {
			return -1;
		}
		int num = 0;
		num += (isStageWall(tile, tStage, level, x - 1, y)) ? 0 : 1;
		num += (isStageWall(tile, tStage, level, x + 1, y)) ? 0 : 1;
		num += (isStageWall(tile, tStage, level, x, y - 1)) ? 0 : 1;
		num += (isStageWall(tile, tStage, level, x, y + 1)) ? 0 : 1;
		return num;
	}

	private boolean isStageWall(byte[] tile, byte[] tStage, int level, int x,
			int y) {
		if (x < 0 || wx <= x || y < 0 || wy <= y) {
			return true;
		} else if (tile[x + y * wx] == -1 || tStage[x + y * wx] > 0) {
			return true;
		}
		return false;
	}

	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append(maxLevel);
		sb.append('\n');
		for (int y = 0; y < wy; y++) {
			for (int x = 0; x < wx; x++) {
				sb.append(stage[x + wx * y]);
			}
			sb.append('\n');
		}
		return sb.toString();
	}

	public int calcScore(World world) {
		int score = 0;
		for (List<List<Integer>> groupIdsList : connectedGroup) {
			boolean allGroupOk = true;
			for (List<Integer> groupIds : groupIdsList) {
				boolean okGroup = true;
				for (Integer groupId : groupIds) {
					int idx = (groupId != world.getTile().length - 1) ? groupId : 0;
					byte v = world.getTile()[idx];
					if (v > 0 && v - 1 == idx) {
						// ok
					} else {
						okGroup = false;
						break;
					}
				}
				if (okGroup) {
					score++;
				} else {
					allGroupOk = false;
				}
			}
			if (!allGroupOk) {
				// 一つのグループが全部OKの場合のみ次のステージへ行ける
				break;
			}
		}
		return score;
	}

	public int getMaxLevel() {
		return maxLevel;
	}

	public int getMaxScore() {
		return maxScore;
	}

	public List<List<List<Integer>>> getConnectedGroup() {
		return connectedGroup;
	}

	private List<List<List<Integer>>> pickConnectedGroup() {
		List<List<List<Integer>>> groupIdxsListList = new ArrayList<List<List<Integer>>>();
		for (int i = 0; i < maxLevel; i++) {
			List<List<Integer>> groupIdxsList = pickConnectedGroup(i + 1);
			groupIdxsListList.add(groupIdxsList);
		}
		return groupIdxsListList;
	}

	private List<List<Integer>> pickConnectedGroup(int level) {
		List<List<Integer>> groupIdxsList = new ArrayList<List<Integer>>();
		boolean[] picked = new boolean[this.stage.length];
		for (int i = 0; i < this.stage.length; i++) {
			if (!picked[i] && this.stage[i] == level) {
				List<Integer> groupIdxs = new ArrayList<Integer>();
				pickConnectedGroup(groupIdxs, picked, level, i % this.wx, i
						/ this.wx);
				groupIdxsList.add(groupIdxs);
			}
		}
		return groupIdxsList;
	}

	private void pickConnectedGroup(List<Integer> result, boolean[] picked,
			int level, int x, int y) {
		int idx = x + y * this.wx;
		if (x < 0 || this.wx <= x || y < 0 || this.wy <= y) {
			return;
		}
		if (this.stage[idx] == level) {
			if (picked[idx]) {
				return;
			} else {
				picked[idx] = true;
				result.add(idx);
				pickConnectedGroup(result, picked, level, x + 1, y);
				pickConnectedGroup(result, picked, level, x, y + 1);
				pickConnectedGroup(result, picked, level, x, y - 1);
				pickConnectedGroup(result, picked, level, x - 1, y);
			}
		}
	}
}
